from nmigen import *

from nap.stream import BasicStream

__all__ = ["stream_transformer"]


def stream_transformer(input_stream: BasicStream, output_stream: BasicStream, m: Module, *, latency: int, handle_out_of_band=True, allow_partial_out_of_band=False):
    """
    A utility to help you writing fixed latency stream ip that converts one input word to one output word.

    :warning:
    You have to make sure that you only sample the input when ready and valid of it are high for transformers with latency
    otherwise you are not going to comply to the stream contract. In this case you MUST place a StreamBuffer after your core.

    @param handle_out_of_band: determines if this core should connect the out of bands signals or if it is done manually
    @param allow_partial_out_of_band: allow the out of band signals of the streams to differ
    @param input_stream: the input stream
    @param output_stream: the output stream
    @param m: a nmigen Module
    @param latency: the latency of the transform data path in cycles
    """
    if latency == 0:
        m.d.comb += output_stream.connect_upstream(input_stream, only=["ready", "valid"])
        if handle_out_of_band:
            if not allow_partial_out_of_band:
                assert list(input_stream.out_of_band_signals.keys()) == list(output_stream.out_of_band_signals.keys())
            for k in input_stream.out_of_band_signals.keys():
                m.d.comb += output_stream[k].eq(input_stream[k])

    elif latency == 1:
        input_transaction = input_stream.ready & input_stream.valid
        output_transaction = output_stream.ready & output_stream.valid

        with m.If(input_transaction):
            if handle_out_of_band:
                if not allow_partial_out_of_band:
                    assert list(input_stream.out_of_band_signals.keys()) == list(output_stream.out_of_band_signals.keys())
                for k in set(input_stream.out_of_band_signals.keys()).intersection(output_stream.out_of_band_signals.keys()):
                    m.d.sync += output_stream[k].eq(input_stream[k])

        output_produce = Signal()
        m.d.sync += output_produce.eq(input_transaction)

        has_output = Signal()
        with m.If(has_output | output_produce):
            m.d.comb += output_stream.valid.eq(1)
        m.d.sync += has_output.eq(has_output + output_produce - output_transaction > 0)
        m.d.comb += input_stream.ready.eq(output_stream.ready | (~has_output & ~output_produce))

    else:
        raise NotImplementedError()


import unittest

import pytest

from nap import SimPlatform, do_nothing
from nap.stream import BasicStream, verify_stream_output_contract, write_to_stream, read_from_stream
from . import UnbufferedAsyncStreamFIFO, BufferedAsyncStreamFIFO, UnbufferedSyncStreamFIFO, \
    BufferedSyncStreamFIFO


class TestFifo(unittest.TestCase):
    def check_fifo_basic(self, fifo_generator):
        input = BasicStream(32)
        fifo = fifo_generator(input, 1024)

        def testbench():
            for i in range(10):
                yield from write_to_stream(input, payload=i)

            # async fifos need some time due to cdc
            yield from do_nothing()

            assert (yield fifo.r_level) == 10

            for i in range(10):
                assert (yield from read_from_stream(fifo.output)) == i, "read data doesnt match written data"

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.sim(fifo, testbench)

    def test_sim_async_stream_fifo(self):
        fifo_gen = lambda input, depth: UnbufferedAsyncStreamFIFO(input, depth, o_domain="sync", i_domain="sync")
        self.check_fifo_basic(fifo_gen)

    def test_async_stream_fifo_buffered(self):
        fifo_gen = lambda input, depth: BufferedAsyncStreamFIFO(input, depth, o_domain="sync", i_domain="sync")
        self.check_fifo_basic(fifo_gen)

    def test_sync_stream_fifo(self):
        fifo_gen = lambda input, depth: UnbufferedSyncStreamFIFO(input, depth)
        self.check_fifo_basic(fifo_gen)

    def test_sync_stream_fifo_buffered(self):
        fifo_gen = lambda input, depth: BufferedSyncStreamFIFO(input, depth)
        self.check_fifo_basic(fifo_gen)

    @pytest.mark.skip("this can not be proven at the moment because a FFSyncronizer in the async FIFO is resetless")
    def test_async_stream_fifo_output_properties(self):
        input = BasicStream(32)
        verify_stream_output_contract(UnbufferedAsyncStreamFIFO(input, 10, o_domain="sync", i_domain="sync"))

    def test_async_stream_fifo_buffered_output_properties(self):
        input = BasicStream(32)
        verify_stream_output_contract(BufferedAsyncStreamFIFO(input, 10, o_domain="sync", i_domain="sync"))

    def test_sync_stream_fifo_output_properties(self):
        input = BasicStream(32)
        verify_stream_output_contract(UnbufferedSyncStreamFIFO(input, 10))

    def test_sync_stream_fifo_buffered_output_properties(self):
        input = BasicStream(32)
        verify_stream_output_contract(BufferedSyncStreamFIFO(input, 10))



from nmigen import *
from nap import StatusSignal, PacketizedStream
from . import ImageStream

__all__ = ["ImageStream2PacketizedStream", "PacketizedStream2ImageStream"]


class ImageStream2PacketizedStream(Elaboratable):
    """Convert an ImageStream to a packetized Stream by producing one packet per frame"""

    def __init__(self, input: ImageStream):
        self.input = input
        self.output = PacketizedStream(self.input.payload.shape(), name="packetized_image_stream")

    def elaborate(self, platform):
        m = Module()

        m.d.comb += self.output.connect_upstream(self.input, exclude=["last", "frame_last", "line_last"])
        m.d.comb += self.output.last.eq(self.input.frame_last)

        return m


class PacketizedStream2ImageStream(Elaboratable):
    """Convert a Packetized stream to an Image stream by creating lines with `width`"""
    def __init__(self, input: PacketizedStream, width):
        self.input = input
        self.width = width
        self.not_exact_number_of_lines_error = StatusSignal(32)
        self.output = ImageStream(input.payload.shape(), name="adapted_image_stream")

    def elaborate(self, platform):
        m = Module()

        line_ctr = Signal(16)

        m.d.comb += self.output.connect_upstream(self.input, only=["ready", "valid", "payload"])
        with m.If(self.input.ready & self.input.valid):
            with m.If(self.input.last):
                m.d.sync += line_ctr.eq(0)
                with m.If(line_ctr != self.width - 1):
                    m.d.sync += self.not_exact_number_of_lines_error.eq(self.not_exact_number_of_lines_error + 1)
            with m.Else():
                with m.If(line_ctr >= (self.width - 1)):
                    m.d.sync += line_ctr.eq(0)
                with m.Else():
                    m.d.sync += line_ctr.eq(line_ctr + 1)

        m.d.comb += self.output.payload.eq(self.input.payload)
        with m.If(self.input.last):
            m.d.comb += self.output.frame_last.eq(1)
            m.d.comb += self.output.line_last.eq(1)
        with m.Else():
            with m.If(line_ctr >= (self.width - 1)):
                m.d.comb += self.output.line_last.eq(1)

        return m



from .stream import *
from .compression import *
from .video import *
from .debug import *
from .axi import *
from .dram_packet_ringbuffer import *
from .jtag import *
from .peripherals import *
from .ila import *
from .cmv12k import *
from .hispi import *
from .mipi import *
from .hdmi import *
from .ft601 import *
from .plugin_module_streamer import *



#!/usr/bin/env python3

from setuptools import setup, find_packages
from pathlib import Path

def scm_version():
    def local_scheme(version):
        if version.tag and not version.distance:
            return version.format_with("")
        else:
            return version.format_choice("", "")
    return {
        "relative_to": __file__,
        "version_scheme": "guess-next-dev",
        "local_scheme": local_scheme
    }


setup(
    name='naps',
    use_scm_version=scm_version(),
    url='https://github.com/apertus-open-source-cinema/nmigen-gateware',
    description='nMigen Apertus Packages',
    long_description=Path("README.rst").read_text(),
    packages=find_packages(),
    install_requires=[
        'huffman',
        'nmigen @ git+https://github.com/nmigen/nmigen.git',
        'nmigen-boards @ git+https://github.com/nmigen/nmigen-boards.git',
    ],
    keywords='nmigen apertus fpga gateware video camera',
    project_urls={
        'Documentation': 'https://apertus-open-source-cinema.github.io/nmigen-gateware/',
        'Homepage': 'https://github.com/apertus-open-source-cinema/nmigen-gateware',
        'Bug Tracker': 'https://github.com/apertus-open-source-cinema/nmigen-gateware/issues'
    },
    extras_require={
        'test': [
            'pytest',
            'pytest-forked',
            'pytest-xdist',
            'psutil',

            'pytest-pycharm',
            'pytest-github-actions-annotate-failures',

            'pydng @ git+https://github.com/schoolpost/PyDNG.git#subdirectory=src',
            'imageio',
            'rawpy',

            'bitarray',

            # needed for python wavelet compressor
            'numba',
            'numpy',

            'scipy',  # only needed for vifp calculation
            'matplotlib',
            'yowasp-yosys',
        ],
        'doc': [
            'sphinx',
            'sphinx_rtd_theme',
        ]
    }
)



from nmigen import *
from nap import Stream, StatusSignal, driver_property

__all__ = ["StreamInfo", "InflexibleSinkDebug", "InflexibleSourceDebug"]


class StreamInfo(Elaboratable):
    def __init__(self, stream: Stream):
        self.stream = stream

        self.reference_counter = StatusSignal(32)
        self.successful_transactions_counter = StatusSignal(32)
        self.ready_not_valid = StatusSignal(32)
        self.valid_not_ready = StatusSignal(32)
        self.current_ready = StatusSignal()
        self.current_valid = StatusSignal()

    def elaborate(self, platform):
        m = Module()

        with m.If(self.stream.valid & ~self.stream.ready):
            m.d.sync += self.valid_not_ready.eq(self.valid_not_ready + 1)
        with m.If(self.stream.ready & ~self.stream.valid):
            m.d.sync += self.ready_not_valid.eq(self.ready_not_valid + 1)

        m.d.sync += self.reference_counter.eq(self.reference_counter + 1)

        transaction = Signal()
        m.d.comb += transaction.eq(self.stream.valid & self.stream.ready)
        with m.If(transaction):
            m.d.sync += self.successful_transactions_counter.eq(self.successful_transactions_counter + 1)

        m.d.comb += self.current_ready.eq(self.stream.ready)
        m.d.comb += self.current_valid.eq(self.stream.valid)

        for name, signal in self.stream.payload_signals.items():
            if len(signal) == 1:
                m.submodules[name] = MetadataSignalDebug(signal, transaction)
            else:
                current_state = StatusSignal(signal.shape())
                m.d.comb += current_state.eq(signal)
                setattr(self, "current_{}".format(name), current_state)
                last_transaction_state = StatusSignal(signal.shape())
                with m.If(transaction):
                    m.d.sync += last_transaction_state.eq(signal)
                setattr(self, "last_transaction_{}".format(name), last_transaction_state)

        return m

    @driver_property
    def efficiency_percent(self):
        return self.successful_transactions_counter / self.reference_counter * 100

    @driver_property
    def stall_source_percent(self):
        return self.ready_not_valid / self.reference_counter * 100

    @driver_property
    def stall_sink_percent(self):
        return self.valid_not_ready / self.reference_counter * 100

    @driver_property
    def stall_both_percent(self):
        return (100 - self.efficiency_percent) - self.stall_source_percent - self.stall_sink_percent


class MetadataSignalDebug(Elaboratable):
    def __init__(self, signal, transaction):
        assert len(signal) == 1
        self.signal = signal
        self.transaction = transaction

        self.cycle_1_length = StatusSignal(32)
        self.cycle_1_length_changed = StatusSignal(32)
        self.cycle_0_length = StatusSignal(32)
        self.cycle_0_length_changed = StatusSignal(32)
        self.cycles = StatusSignal(32)
        self.current = StatusSignal()

    def elaborate(self, platform):
        m = Module()

        m.d.comb += self.current.eq(self.signal)

        cycle_0_counter = Signal(32)
        cycle_1_counter = Signal(32)

        with m.If(self.transaction):
            with m.If(self.signal):
                m.d.sync += cycle_0_counter.eq(0)
                with m.If(cycle_0_counter != 0):
                    m.d.sync += self.cycle_0_length.eq(cycle_0_counter)
                    with m.If(self.cycle_0_length != cycle_0_counter):
                        m.d.sync += self.cycle_0_length_changed.eq(self.cycle_0_length_changed + 1)
                m.d.sync += cycle_1_counter.eq(cycle_1_counter + 1)
            with m.Else():
                m.d.sync += cycle_1_counter.eq(0)
                with m.If(cycle_1_counter != 0):
                    m.d.sync += self.cycle_1_length.eq(cycle_1_counter)
                    with m.If(self.cycle_1_length != cycle_1_counter):
                        m.d.sync += self.cycle_1_length_changed.eq(self.cycle_1_length_changed + 1)
                m.d.sync += cycle_0_counter.eq(cycle_0_counter + 1)

            signal_last = Signal()
            m.d.sync += signal_last.eq(self.signal)
            with m.If(self.signal & ~signal_last):
                m.d.sync += self.cycles.eq(self.cycles + 1)

        return m

    @driver_property
    def cycle_length(self):
        return self.cycle_0_length + self.cycle_1_length


class InflexibleSourceDebug(Elaboratable):
    def __init__(self, stream):
        self.stream = stream
        self.dropped = StatusSignal(32)

    def elaborate(self, platform):
        m = Module()

        with m.If(self.stream.valid & ~self.stream.ready):
            m.d.sync += self.dropped.eq(self.dropped + 1)

        return m


class InflexibleSinkDebug(Elaboratable):
    def __init__(self, stream):
        self.stream = stream
        self.invalid = StatusSignal(32)

    def elaborate(self, platform):
        m = Module()

        with m.If(self.stream.ready & ~self.stream.valid):
            m.d.sync += self.invalid.eq(self.invalid + 1)

        return m



import re

__all__ = ["decimal_range", "camel_to_snake"]


def decimal_range(start, stop, step):
    next_decimal = start
    while next_decimal < stop:
        yield next_decimal
        next_decimal += step
    return


def camel_to_snake(name):
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()



from nmigen import *

from nap import ControlSignal, StatusSignal
from nap.cores import InflexibleSinkDebug, StreamInfo, ImageStream, RGB24
from .hdmi_tx import HdmiTx

__all__ = ['HdmiStreamSink', 'HdmiStreamAligner']


class HdmiStreamSink(Elaboratable):
    def __init__(self, input: ImageStream, resource, modeline, pix_domain="pix", generate_clocks=True):
        self.input = input
        self.generate_clocks = generate_clocks
        self.pix_domain = pix_domain
        self.modeline = modeline
        self.resource = resource

    def elaborate(self, platform):
        m = Module()

        hdmi = m.submodules.hdmi = HdmiTx(self.resource, self.modeline, self.pix_domain, self.generate_clocks)
        m.submodules.aligner = DomainRenamer(self.pix_domain)(HdmiStreamAligner(self.input, hdmi))
        m.d.comb += hdmi.rgb.eq(RGB24(self.input.payload))

        return m


class HdmiStreamAligner(Elaboratable):
    """
    Aligns the HDMI output to the Image stream by 'slipping' data during the blanking periods until the frame is
    aligned.
    """
    def __init__(self, input: ImageStream, hdmi):
        self.hdmi = hdmi
        self.input = input

        self.allow_slip_h = ControlSignal(reset=1)
        self.allow_slip_v = ControlSignal(reset=1)
        self.slipped_v = StatusSignal(32)
        self.slipped_h = StatusSignal(32)

        self.line_cycles = StatusSignal(32)
        self.frame_cycles = StatusSignal(32)

    def elaborate(self, platform):
        m = Module()

        m.submodules.debug = InflexibleSinkDebug(self.input)
        input_stream_info = m.submodules.input_stream_info = StreamInfo(stream=self.input)

        line_length_counter = Signal(32)
        was_blanking_x = Signal()
        m.d.sync += was_blanking_x.eq(self.hdmi.timing_generator.is_blanking_x)
        with m.If(~self.hdmi.timing_generator.is_blanking_x):
            m.d.sync += line_length_counter.eq(line_length_counter + 1)
        with m.If(self.hdmi.timing_generator.is_blanking_x & ~was_blanking_x):
            m.d.sync += self.line_cycles.eq(line_length_counter)
            m.d.sync += line_length_counter.eq(0)

        frame_length_counter = Signal(32)
        was_blanking_y = Signal()
        m.d.sync += was_blanking_y.eq(self.hdmi.timing_generator.is_blanking_y)
        with m.If(self.hdmi.timing_generator.active):
            m.d.sync += frame_length_counter.eq(frame_length_counter + 1)
        with m.If(self.hdmi.timing_generator.is_blanking_y & ~was_blanking_y):
            m.d.sync += self.frame_cycles.eq(frame_length_counter)
            m.d.sync += frame_length_counter.eq(0)

        with m.If(self.hdmi.timing_generator.active):
            m.d.comb += self.input.ready.eq(1)

        was_line_last = Signal()
        was_frame_last = Signal()
        with m.If(self.input.ready):
            m.d.sync += was_line_last.eq(self.input.line_last)
            m.d.sync += was_frame_last.eq(self.input.frame_last)

        with m.If(self.hdmi.timing_generator.is_blanking_x & ~was_line_last & self.allow_slip_h):
            m.d.sync += self.slipped_h.eq(self.slipped_h + 1)
            m.d.comb += self.input.ready.eq(1)

        with m.If(self.hdmi.timing_generator.is_blanking_y & ~was_frame_last & self.allow_slip_v):
            m.d.sync += self.slipped_v.eq(self.slipped_v + 1)
            m.d.comb += self.input.ready.eq(1)

        return m



# TODO: add tests

from nmigen import *
from nap.soc import SocPlatform, Response


__all__ = ["DrpInterface", "DrpBridge"]


# TODO: let this be an `Interface`
class DrpInterface:
    def __init__(self, DWE, DEN, DADDR, DI, DO, DRDY, DCLK):
        self.data_write_enable: Signal = DWE
        self.data_enable: Signal = DEN
        self.address: Signal = DADDR
        self.data_in: Signal = DI
        self.data_out: Signal = DO
        self.ready: Signal = DRDY
        self.clk : Signal = DCLK


class DrpBridge(Elaboratable):
    def __init__(self, drp_interface):
        """
        A bridge for the xilinx dynamic reconfiguration port. This is for example used in the Xilinx 7 series MMCM and
        PLL vendor.

        :param drp_interface: the drp bus of the drp slave
        """
        self.drp_interface: DrpInterface = drp_interface

    def elaborate(self, platform: SocPlatform):
        m = Module()

        def handle_read(m, addr, data, read_done):
            m.d.comb += self.drp_interface.clk.eq(ClockSignal())
            m.d.sync += self.drp_interface.address.eq(addr)
            m.d.sync += self.drp_interface.data_enable.eq(1)
            with m.If(self.drp_interface.ready):
                m.d.sync += self.drp_interface.data_enable.eq(0)
                m.d.sync += data.eq(self.drp_interface.data_out)
                read_done(Response.OK)

        def handle_write(m, addr, data, write_done):
            m.d.comb += self.drp_interface.clk.eq(ClockSignal())
            m.d.sync += self.drp_interface.address.eq(addr)
            m.d.sync += self.drp_interface.data_enable.eq(1)
            m.d.sync += self.drp_interface.data_write_enable.eq(1)
            m.d.sync += self.drp_interface.data_in.eq(data)
            with m.If(self.drp_interface.ready):
                m.d.sync += self.drp_interface.data_enable.eq(0)
                m.d.sync += self.drp_interface.data_write_enable.eq(0)
                write_done(Response.OK)

        # TODO: fix drp bridge
        # memorymap = MemoryMap()
        # memorymap.allocate("drp", writable=True, bits=2**len(self.drp_interface.address) * 8)
        #
        # m.submodules += Peripheral(
        #     handle_read,
        #     handle_write,
        #     memorymap
        # )

        return m



from nmigen import *


def _opposed_direction(dir):
    return { "input": "output", "output": "input", "bidirectional": "bidirectional" }[dir]


def _dir_to_netlistsvg(dir):
    return { "i": "input", "o": "output", "io": "bidirectional" }[dir]


def _beautify_nmigen_name(s):
    return s.replace("__", "/")


def _hierarchy_to_dot_frag(frag: Fragment, hierarchy = ["top"], _bits = [0]):
    def next_bit():
        nonlocal _bits
        bit = _bits[0]
        _bits[0] += 1

        return bit

    from nmigen.hdl import ir

    cells = {}

    cell = {}
    cell["type"] = hierarchy[-1]
    cell["port_directions"] = {}
    cell["connections"] = {}

    ports = {}

    if isinstance(frag, ir.Instance):
        for name, (value, dir) in frag.named_ports.items():
            dir = _dir_to_netlistsvg(dir)

            ports[name] = {}
            ports[name]["direction"] = dir

            cell["port_directions"][name] = dir
            cell["connections"][name] = [next_bit() for _ in range(value._shape().width)]
    else:
        ports = {}

        for sig, dir in frag.ports.items():
            name =_beautify_nmigen_name(sig.name)
            dir = _dir_to_netlistsvg(dir)

            ports[name] = {}
            ports[name]["direction"] = dir
            cell["port_directions"][name] = dir
            cell["connections"][name] = [next_bit() for _ in range(sig._shape().width)]

        for (sub_frag, sub_name) in frag.subfragments:
            if sub_name == "instance":
                sub_name = hierarchy[-1] + "/" + sub_name

            new_cells, new_ports = _hierarchy_to_dot_frag(sub_frag, hierarchy + [sub_name])

            # print("hierarchy", hierarchy)
            # print("submod", sub_name)
            # print("my ports", ports)
            # print("new_ports", new_ports)
            # print("new_cells", new_cells)
            # print("cell[port_directions]", cell["port_directions"])


            # print(ports)

            for port_name, dir in new_ports.items():
                if port_name in cell["port_directions"]:
                    wrapped_port_name = sub_name + "/" + port_name
                else:
                    wrapped_port_name = port_name

                assert wrapped_port_name not in cell["port_directions"]

                dir = dir["direction"]
                cell["port_directions"][wrapped_port_name] = _opposed_direction(dir)
                cell["connections"][wrapped_port_name] = new_cells[sub_name]["connections"][port_name]

            for name, cell_ in new_cells.items():
                if name not in cells:
                    cells[name] = cell_
                else:
                    i = 1

                    while name in cells:
                        name = hierarchy[-i] + "/" + sub_name
                        i += 1

                    cells[name] = cell_

        # print(frag.named_ports)
        # print(frag.subfragments)


    cells[hierarchy[-1]] = cell

    return cells, ports


def hierarchy_to_dot(e: Elaboratable, plat = None, **kwargs):
    from nmigen.hdl import ir
    frag = ir.Fragment.get(e, plat).prepare()

    cells, toplevel_ports = _hierarchy_to_dot_frag(frag)

    for port_name in list(toplevel_ports.keys()):
        wrapped_port_name = port_name.replace(".", "/")

        if wrapped_port_name != port_name:
            toplevel_ports[wrapped_port_name] = toplevel_ports[port_name]
            del toplevel_ports[port_name]

        toplevel_ports[wrapped_port_name]["bits"] = cells["top"]["connections"][port_name]


    out = { "modules": { "top": { "cells": cells, "ports": toplevel_ports } } }

    import json
    return json.dumps(out, indent=4)


from nmigen import *
from nmigen.hdl.ast import UserValue
from nap.util import yosys


def InstanceHelper(source_files, instance_name):
    return type(instance_name, (HelperedInstance,), {"instance_name": instance_name, "source_files": source_files})


class HelperedInstance(Elaboratable):
    def __init__(self, **kwargs):
        self.parameters = kwargs
        self.ports = yosys.get_verilog_module_ports(self.source_files, self.instance_name)
        self.signal_proxy = SignalProxy(self.ports)

    def __call__(self, *args, **kwargs):
        assert len(args) == 0
        self.parameters.update(**kwargs)
        return self

    def __getitem__(self, item):
        return self.signal_proxy.__getitem__(item)

    def __getattr__(self, item):
        return self.signal_proxy.__getattr__(item)

    def elaborate(self, platform):
        m = Module()

        named_ports = {
            "{}_{}".format(self.ports[name]["direction"], name): signal
            for name, signal in self.signal_proxy.used_ports.items()
        }

        def legalize_parameter(parameter):
            if isinstance(parameter, bool):
                return "{}".format(parameter).upper()
            elif isinstance(parameter, str):
                return parameter.upper()
            else:
                return parameter

        legalized_parameters = {k: legalize_parameter(v) for k, v in self.parameters.items()}
        parameters = {"p_{}".format(k.upper()): v for k, v in legalized_parameters.items()}
        m.submodules.instance = Instance(self.instance_name, **named_ports, **parameters)

        return m


class SignalProxy(UserValue):
    def __init__(self, ports, path=""):
        super().__init__()

        self.ports = ports
        self.path = path

        self._used_ports = {}
        self.children = {}

    def lower(self):
        if self.path in self.ports:
            if self.path not in self._used_ports:
                self._used_ports[self.path] = Signal(self.ports[self.path]["width"], name=self.path)
            return self._used_ports[self.path]
        else:
            raise PortNotFoundException()

    @property
    def used_ports(self):
        child_ports = {}
        for c in self.children.values():
            child_ports = {**child_ports, **c.used_ports}
        return {**self._used_ports, **child_ports}

    def __getitem__(self, item):
        try:
            return self.lower()[item]
        except PortNotFoundException:
            return self.__getattr__(str(item))

    def __getattr__(self, item):
        if item.startswith("__"):
            return

        if item == "in_":
            item = "in"

        item = item.upper()

        if item not in self.children:
            self.children[item] = SignalProxy(self.ports, path=(self.path + item))
        return self.children[item]


class PortNotFoundException(Exception):
    pass



import unittest
from nap.stream import PacketizedStream, LegalStreamSource, verify_stream_output_contract
from .tee import StreamTee, StreamCombiner


class StreamTeeTest(unittest.TestCase):
    def test_tee_output_stream_contract(self):
        input_stream = PacketizedStream(32)
        dut = StreamTee(input_stream)
        output1, output2 = dut.get_output(), dut.get_output()
        verify_stream_output_contract(dut, stream_output=output1, support_modules=(LegalStreamSource(input_stream),))
        verify_stream_output_contract(dut, stream_output=output2, support_modules=(LegalStreamSource(input_stream),))

    def test_stream_combiner_output_stream_contract(self):
        input1, input2 = PacketizedStream(32), PacketizedStream(32)
        dut = StreamCombiner(input1, input2, merge_payload=True)
        verify_stream_output_contract(dut, support_modules=(LegalStreamSource(input1), LegalStreamSource(input2)))



from functools import lru_cache
from json import loads

from nmigen._toolchain.yosys import find_yosys


@lru_cache()
def parse_yosys_json(verilog_paths):
    if isinstance(verilog_paths, str):
        verilog_paths = [verilog_paths]
    json = yosys_script([
        "\n".join("read_verilog {}".format(path) for path in verilog_paths),
        "write_json"
    ])
    return loads(json)


def get_verilog_module_ports(verilog_paths, module_name):
    """Get the ports of a verilog module via yosys

    :param verilog_paths:
    :param module_name:
    :return: the verilog as a string
    """

    parsed_json = parse_yosys_json(verilog_paths)
    module = parsed_json["modules"][module_name]
    ports = module["ports"]

    # do some reformatting of the data:
    def abbrev_direction(long_form):
        if long_form in ("output", "o"):
            return "o"
        elif long_form in ("input", "i"):
            return "i"
        elif long_form in ("inout", "io"):
            return "io"
        else:
            raise Exception("Bad direction: {}".format(long_form))

    for k, v in ports.items():
        if "bits" in v:
            ports[k]["width"] = len(v["bits"])
            del ports[k]["bits"]
        assert "width" in v

        ports[k]["direction"] = abbrev_direction(v["direction"])

    return ports


def yosys_script(commands):
    """Executes a yosys script

    :param commands: a list of commands to run
    :return: the stdout of yosys
    """

    return find_yosys(lambda ver: ver >= (0, 9)).run(["-q", "-p {}".format(";\n".join(commands))])



from typing import List
from nmigen import *
from nmigen.lib.io import Pin
from nap import PacketizedFirstStream, StreamGearbox, nAny, nAll
from .s7_rx_phy import MipiClockRxPhy
from .aligner import LaneWordAligner


class MipiLaneCombiner(Elaboratable):
    """
    Combines 1 to 4 lanes to a 32 bit word that can be used for parsing the packet header.
    Also assists with the training of multiple lanes.
    """

    def __init__(self, clock_pin: Pin, data_pins: List[Pin], domain_prefix="mipi"):
        self.domain_prefix = domain_prefix
        self.clock_pin = clock_pin
        self.data_pins = data_pins

        self.in_packet = Signal()  # input; indicates if the upper layer is in a valid packet
        self.output = PacketizedFirstStream(32)

    def elaborate(self, platform):
        m = Module()

        word_domain = f'{self.domain_prefix}_word'
        ddr_domain = f'{self.domain_prefix}_ddr'

        clock_phy = m.submodules.clock_phy = MipiClockRxPhy(self.clock_pin, ddr_domain)
        lane_phys = [LaneWordAligner(pin, ddr_domain, self.in_packet) for pin in self.data_pins]
        for i, phy in enumerate(lane_phys):
            m.submodules[f'lane_{i + 1}_phy'] = phy

        # training of multiple lanes:
        # if some (but not all) lanes assert maybe_first_packet_byte we disable the training logic for the ones that asserted maybe_first_packet_byte
        # and let the others continue their Training. This however can lead to endless loops (when some lane is trained on not the sync pattern but some
        # other piece of data). To reduce the likeliness that this happens we only re-enable the training logic if the training did not work after a timeout
        # (measured in assertions of maybe_first_packet_byte)
        timeout = 27
        timeout_ctr = Signal(range(timeout))

        def reset_successive_training():
            m.d.sync += timeout_ctr.eq(0)
            for l in lane_phys:
                m.d.sync += l.enable_train_logic.eq(1)

        mfpb = [l.maybe_first_packet_byte for l in lane_phys]
        with m.If(~self.in_packet & nAll(mfpb)):  # we completed training successfully
            reset_successive_training()
        with m.Elif(~self.in_packet & nAny(mfpb) & ~nAll(l.maybe_first_packet_byte | l.enable_train_logic for l in lane_phys)):  # we are certainly not on the right path
            reset_successive_training()
        with m.Elif(~self.in_packet & nAny(mfpb) & (timeout_ctr < timeout)):
            m.d.sync += timeout_ctr.eq(timeout_ctr + 1)
            for l in lane_phys:
                with m.If(l.maybe_first_packet_byte & nAny(l.maybe_first_packet_byte & ~l.enable_train_logic for l in lane_phys)):
                    m.d.sync += l.enable_train_logic.eq(0)
        with m.Elif(~self.in_packet & nAny(mfpb)):
            reset_successive_training()

        gearbox_input = PacketizedFirstStream(8 * len(lane_phys))
        m.d.comb += gearbox_input.first.eq(nAll(mfpb))
        for i, l in enumerate(lane_phys):
            m.d.comb += gearbox_input.payload[i * 8: (i + 1) * 8].eq(l.output)
        gearbox = m.submodules.gearbox = StreamGearbox(gearbox_input, target_width=32)
        m.d.comb += self.output.connect_upstream(gearbox.output)

        return m



from abc import ABC

from .fatbitstream import FatbitstreamContext
from .hooks import csr_and_driver_method_hook, address_assignment_hook, peripherals_collect_hook
from .pydriver.generate import pydriver_hook
from .tracing_elaborate import fragment_get_with_elaboratable_trace

__all__ = ["SocPlatform"]


class SocPlatform(ABC):
    base_address = None
    _wrapped_platform = None

    # we build a new type that combines the soc and the real platform class
    def __new__(cls, platform, *args, **kwargs):
        return super(SocPlatform, cls).__new__(type(cls.__name__, (cls, platform.__class__), vars(platform)))

    # we pass through all platform methods, because we pretend to be a platform
    def __getattr__(self, item):
        return getattr(self._wrapped_platform, item)

    def __init__(self, platform):
        self._wrapped_platform = platform

        # inject our prepare method into the platform as a starting point for all our hooks
        self.real_prepare = self._wrapped_platform.prepare
        self._wrapped_platform.prepare = self.prepare

        # inject fatbitstream generation into the platform templates
        # we it this way because command_templates might be a property object that cant be written to directly
        original_command_templates = self._wrapped_platform.command_templates
        self._wrapped_platform.extra_command_templates = []
        self._wrapped_platform.__class__.command_templates = property(lambda plat: [
            *original_command_templates,
            *plat.extra_command_templates
        ])

        # store a reference in the platform that is wrapped to be able to retrieve it during e.g. fatbitstream
        # generation
        self._wrapped_platform._soc_platform = self

        self.prepare_hooks = []
        self.to_inject_subfragments = []
        self.final_to_inject_subfragments = []

        self.prepare_hooks.append(csr_and_driver_method_hook)
        self.prepare_hooks.append(address_assignment_hook)
        self.prepare_hooks.append(peripherals_collect_hook)
        self.prepare_hooks.append(pydriver_hook)

    # we override the prepare method of the real platform to be able to inject stuff into the design
    def prepare(self, elaboratable, name="top", *args, **kwargs):
        print("# ELABORATING MAIN DESIGN")
        top_fragment, sames = fragment_get_with_elaboratable_trace(elaboratable, self)

        def inject_subfragments(top_fragment, sames, to_inject_subfragments):
            for elaboratable, name in to_inject_subfragments:
                fragment, fragment_sames = fragment_get_with_elaboratable_trace(elaboratable, self, sames)
                print("<- injecting fragment '{}'".format(name))
                top_fragment.add_subfragment(fragment, name)
            self.to_inject_subfragments = []

        print("\n# ELABORATING SOC PLATFORM ADDITIONS")
        inject_subfragments(top_fragment, sames, self.to_inject_subfragments)
        for hook in self.prepare_hooks:
            print("-> running {}".format(hook.__name__))
            hook(self, top_fragment, sames)
            inject_subfragments(top_fragment, sames, self.to_inject_subfragments)

        print("\ninjecting final fragments")
        inject_subfragments(top_fragment, sames, self.final_to_inject_subfragments)

        print("\ninjecting fatbitstream generation code")
        fc = FatbitstreamContext.get(self)
        self._wrapped_platform.extra_command_templates.extend(fc.generate_fatbitstream_generator(name))

        print("\n\nexiting soc code\n")

        return self.real_prepare(top_fragment, name, *args, **kwargs)



from pprint import pprint
from nmigen import *
from nap import StatusSignal

__all__ = ["Pll", "EClkSync", "ClkDiv", "Osc"]


class Pll(Elaboratable):
    vco_multipliers = list(range(2, 128))
    vco_dividers = list(range(1, 128))
    output_dividers = list(range(1, 128))

    @staticmethod
    def is_valid_vco_conf(input_freq, mul, div, exception=False):
        if not mul in Pll.vco_multipliers:
            if exception:
                raise ValueError
            return False
        if not div in Pll.vco_dividers:
            if exception:
                raise ValueError
            return False
        vco_freq = input_freq * mul / div
        if 200e6 > vco_freq:
            if exception:
                raise ValueError(vco_freq)
            return False
        if 800e6 < vco_freq:
            if exception:
                raise ValueError(vco_freq)
            return False
        return True

    def __init__(self, input_freq, vco_mul, vco_div, input_domain="sync"):
        Pll.is_valid_vco_conf(input_freq, vco_mul, vco_div, exception=True)
        self.vco_freq = input_freq * vco_mul / vco_div

        self.locked = StatusSignal()

        self.m = Module()

        clkop = Signal()
        clkfb = Signal()
        self.m.submodules.clkbufa = Instance(
            "CLKFBBUFA",
            i_A=clkop,
            o_Z=clkfb
        )

        self.output_port_names = ["CLKOS", "CLKOS2", "CLKOS3"]

        self.params = {
            "CLKI_DIV": vco_div,

            # we use OLKOP as our feedback path
            "CLKFB_DIV": vco_mul,
            "FEEDBK_PATH": "USERCLOCK",
            "CLKOP_DIV": 1,
            "CLKOP_ENABLE": "ENABLED",

            **{"{}_ENABLE".format(output): "DISABLED" for output in self.output_port_names}
        }
        self.attributes = {
            "FREQUENCY_PIN_CLKI": str(input_freq / 1e6),
            "FREQUENCY_PIN_CLKOP": str(self.vco_freq / 1e6),
        }

        self.clocks = []  # (signal, frequency, domain)

        self.clocks.append((clkop, self.vco_freq, None))
        self.inputs = {
            "CLKI": ClockSignal(input_domain),
            "RST": ResetSignal(input_domain),
            "CLKFB": clkfb,
        }
        self.outputs = {
            "LOCK": self.locked,
            "CLKOP": clkop,
        }

    def output_domain(self, domain_name, divisor):
        m = self.m

        cd = ClockDomain(domain_name)
        m.domains += cd
        self.clocks.append((cd.clk, self.vco_freq / divisor, domain_name))
        m.d.comb += ResetSignal(domain_name).eq(~self.locked)

        output_name = self.output_port_names.pop(0)
        self.outputs[output_name] = ClockSignal(domain_name)
        self.params["{}_DIV".format(output_name)] = divisor
        self.params["{}_ENABLE".format(output_name)] = "ENABLED"
        self.attributes["FREQUENCY_PIN_{}".format(output_name)] = str(self.vco_freq / divisor / 1e6)

    def elaborate(self, platform):
        m = self.m
        pprint({**{"p_{}".format(k): v for k, v in self.params.items()},
            **{"i_{}".format(k): v for k, v in self.inputs.items()},
            **{"o_{}".format(k): v for k, v in self.outputs.items()},})
        inst = m.submodules.inst = Instance(
            "EHXPLLJ",
            **{"p_{}".format(k): v for k, v in self.params.items()},
            **{"i_{}".format(k): v for k, v in self.inputs.items()},
            **{"o_{}".format(k): v for k, v in self.outputs.items()},
        )
        inst.attrs.update(self.attributes)

        for signal, frequency, domain in self.clocks:
            if hasattr(platform, "add_clock_constraint"):
                platform.add_clock_constraint(signal, frequency)
            elif domain is not None:
                platform.add_sim_clock(domain, frequency)

        return m


class EClkSync(Elaboratable):
    def __init__(self, input_domain, output_domain, input_frequency=None):
        self.input_frequency = input_frequency
        self.input_domain = input_domain
        self.output_domain = output_domain

    def elaborate(self, platform):
        m = Module()

        cd_output = ClockDomain(self.output_domain)
        m.domains += cd_output
        m.d.comb += ResetSignal(self.output_domain).eq(ResetSignal(self.input_domain))
        if self.input_frequency is not None:
            platform.add_clock_constraint(cd_output.clk, self.input_frequency)

        m.submodules.inst = Instance(
            "ECLKSYNCA",

            i_ECLKI=ClockSignal(self.input_domain),
            i_STOP=Const(0),
            o_ECLKO=ClockSignal(self.output_domain),
        )

        return m


class ClkDiv(Elaboratable):
    def __init__(self, input_domain, output_div_domain, div, bitslip=None, output_x1_domain=None, input_frequency=None):
        self.div = div
        self.output_div_domain = output_div_domain
        self.output_x1_domain = output_x1_domain
        self.input_frequency = input_frequency
        self.input_domain = input_domain
        self.bitslip = bitslip

    def elaborate(self, platform):
        m = Module()

        cd_div = ClockDomain(self.output_div_domain)
        m.domains += cd_div
        m.d.comb += ResetSignal(self.output_div_domain).eq(ResetSignal(self.input_domain))

        additional = {}
        if self.output_x1_domain is not None:
            cd_x1 = ClockDomain(self.output_x1_domain)
            m.domains = cd_x1
            m.d.comb += ResetSignal(self.output_x1_domain).eq(ResetSignal(self.input_domain))
            additional["o_CDIV1"] = ClockSignal(self.output_x1_domain)
            if self.input_frequency is not None:
                platform.add_clock_constraint(cd_x1.clk, self.input_frequency)
        if self.bitslip is not None:
            additional["i_ALIGNWD"] = self.bitslip

        if self.input_frequency is not None:
            platform.add_clock_constraint(cd_div.clk, self.input_frequency / self.div)

        m.submodules.inst = Instance(
            "CLKDIVC",
            p_DIV=str(float(self.div)),

            i_CLKI=ClockSignal(self.input_domain),
            o_CDIVX=ClockSignal(self.output_div_domain),
            **additional
        )

        return m


class Osc(Elaboratable):
    def __init__(self, output_domain="sync", freq=2.08e6):
        self.output_domain = output_domain
        self.freq = freq

    def elaborate(self, platform):
        m = Module()

        m.domains += ClockDomain(self.output_domain)
        inst = m.submodules.inst = Instance(
            "OSCH",

            o_OSC=ClockSignal(self.output_domain),
        )
        inst.attrs["NOM_FREQ"] = str(self.freq / 1e6)

        return m



import unittest
from nmigen import *
from nap import SimPlatform
from nap.cores.axi import axi_ram_sim_model
from nap.stream import PacketizedStream, write_packet_to_stream, read_packet_from_stream
from . import DramPacketRingbufferCpuReader, DramPacketRingbufferStreamWriter, DramPacketRingbufferStreamReader


class StreamIfTest(unittest.TestCase):
    def test_integration(self):
        plat = SimPlatform()
        m = Module()
        writer_axi_port, reader_axi_port = axi_ram_sim_model(plat)

        input_stream = PacketizedStream(64)
        writer = m.submodules.writer = DramPacketRingbufferStreamWriter(input_stream, base_address=0, max_packet_size=10000, n_buffers=4, axi=writer_axi_port)
        reader = m.submodules.reader = DramPacketRingbufferStreamReader(writer, axi=reader_axi_port)
        cpu_reader = m.submodules.cpu_reader = DramPacketRingbufferCpuReader(writer)

        def testbench():
            test_packets = [
                [0 for _ in range(100)],
                [i for i in range(100)]
            ]

            for p in test_packets:
                yield from write_packet_to_stream(input_stream, p)

            read_packets = []
            while len(read_packets) < len(test_packets):
                read = yield from read_packet_from_stream(reader.output)
                read_packets.append(read)

            self.assertEqual(read_packets, test_packets)
            self.assertEqual((yield writer.buffer_level_list[0]), 800)
            self.assertEqual((yield writer.buffer_level_list[1]), 800)

            self.assertEqual((yield cpu_reader.buffer0_level), 800)
            self.assertEqual((yield cpu_reader.buffer1_level), 800)

        plat.add_sim_clock("sync", 100e6)
        plat.sim(m, testbench)



from os.path import join, dirname

from nmigen import Fragment, Module, ClockSignal, DomainRenamer

from ... import *

__all__ = ["ZynqSocPlatform"]


class ZynqSocPlatform(SocPlatform):
    base_address = Address(0x4000_0000, 0, (0x7FFF_FFFF - 0x4000_0000) * 8)
    pydriver_memory_accessor = open(join(dirname(__file__), "memory_accessor_devmem.py")).read()

    def __init__(self, platform, use_axi_interconnect=False):
        from nap.vendor.xilinx_s7 import PS7

        super().__init__(platform)
        self.ps7 = PS7(here_is_the_only_place_that_instanciates_ps7=True)
        self.final_to_inject_subfragments.append((self.ps7, "ps7"))

        def peripherals_connect_hook(platform, top_fragment: Fragment, sames):
            from nap.cores.axi import AxiEndpoint, AxiLitePeripheralConnector, AxiFullToLiteBridge, AxiInterconnect

            if platform.peripherals:
                m = Module()
                platform.ps7.fck_domain(domain_name="axi_lite", requested_frequency=10e6)
                if not hasattr(platform, "is_sim"):  # we are not in a simulation platform
                    axi_full_port: AxiEndpoint = platform.ps7.get_axi_gp_master(ClockSignal("axi_lite"))
                    axi_lite_bridge = m.submodules.axi_lite_bridge = DomainRenamer("axi_lite")(
                        AxiFullToLiteBridge(axi_full_port)
                    )
                    axi_lite_master = axi_lite_bridge.lite_master
                else:  # we are in a simulation platform
                    axi_lite_master = AxiEndpoint(addr_bits=32, data_bits=32, lite=True)
                    self.axi_lite_master = axi_lite_master

                if use_axi_interconnect:
                    interconnect = m.submodules.interconnect = DomainRenamer("axi_lite")(
                        AxiInterconnect(axi_lite_master)
                    )
                    for peripheral in platform.peripherals:
                        controller = DomainRenamer("axi_lite")(AxiLitePeripheralConnector(peripheral))
                        m.d.comb += interconnect.get_port().connect_downstream(controller.axi)
                        m.submodules += controller
                else:
                    aggregator = PeripheralsAggregator()
                    for peripheral in platform.peripherals:
                        aggregator.add_peripheral(peripheral)
                    controller = DomainRenamer("axi_lite")(AxiLitePeripheralConnector(aggregator))
                    m.d.comb += axi_lite_master.connect_downstream(controller.axi)
                    m.submodules += controller
                platform.to_inject_subfragments.append((m, "axi_lite"))
        self.prepare_hooks.append(peripherals_connect_hook)

    def pack_bitstream_fatbitstream(self, builder):
        self.add_file("to_raw_bitstream.py", open(join(dirname(__file__), "to_raw_bitstream.py")).read())
        builder.append_host("python3 to_raw_bitstream.py {{name}}.bit > {{name}}.bin")
        builder.append_self_extracting_blob_from_file("{{name}}.bin", "/usr/lib/firmware/{{name}}.bin")
        builder.append_command("echo {{name}}.bin > /sys/class/fpga_manager/fpga0/firmware\n")

    def toolchain_program(self, *args, **kwargs):
        program_bitstream_ssh(self, *args, **kwargs)



# utility functions for generating and loading devicetree overlays
from textwrap import dedent, indent

from nmigen import Fragment
from . import FatbitstreamContext

__all__ = ["devicetree_overlay"]


def devicetree_overlay(platform, overlay_name, overlay_content, placeholder_substitutions_dict=None):
    if placeholder_substitutions_dict is None:
        placeholder_substitutions_dict = {}

    if not hasattr(platform, "devicetree_overlays"):
        platform.devicetree_overlays = dict()

        def overlay_hook(platform, top_fragment: Fragment, sames):
            assert hasattr(top_fragment, "memorymap")
            memorymap = top_fragment.memorymap

            for overlay_name, (overlay_content, placeholder_substitutions_dict) in reversed(platform.devicetree_overlays.items()):
                things_to_replace = {k: v if isinstance(v, str) else "0x{:x}".format(memorymap.find_recursive(v).address)
                                     for k, v in
                                     placeholder_substitutions_dict.items()}
                things_to_replace["overlay_name"] = overlay_name

                formatted_overlay_text = dedent(overlay_content)
                for name, replacement in things_to_replace.items():
                    formatted_overlay_text = formatted_overlay_text.replace("%{}%".format(name), replacement)

                overlay_text = dedent("""
                    /dts-v1/;
                    /plugin/;
    
                    / {
                        fragment@0 {
                            target = <&amba>;
    
                            __overlay__ {
                                %s
                            };
                        };
                    };
                """ % indent(formatted_overlay_text, "                                "))
                print(overlay_text)

                fc = FatbitstreamContext.get(platform)
                fc.self_extracting_blobs["{}_overlay.dts".format(overlay_name)] = overlay_text
                fc.init_commands.append("mkdir -p /sys/kernel/config/device-tree/overlays/{}\n".format(overlay_name))
                fc.init_commands.append(
                    "dtc -O dtb -@ {0}_overlay.dts -o - > /sys/kernel/config/device-tree/overlays/{0}/dtbo\n\n"
                        .format(overlay_name)
                )
        platform.prepare_hooks.append(overlay_hook)

    assert overlay_name not in platform.devicetree_overlays.keys()
    platform.devicetree_overlays[overlay_name] = (overlay_content, placeholder_substitutions_dict)

    return overlay_name



from nmigen.build import *
from nmigen_boards.te0714_03_50_2I import TE0714_03_50_2IPlatform

__all__ = ["HdmiDigitizerPlatform"]


class HdmiDigitizerPlatform(TE0714_03_50_2IPlatform):
    def __init__(self):
        super().__init__()
        self.add_resources([
            Resource("ft601", 0,
                Subsignal("reset", PinsN("39", dir="o", conn=("JM2", 0))),
                Subsignal("data", Pins("34 32 30 28 26 24 22 20 16 14 12 10 8 6 4 2"
                                       " 1 3 5 7 9 11 13 15 19 21 23 25 27 29 31 33",
                                       dir="io", conn=("JM2", 0))),
                Subsignal("be", Pins("45 51 44 48", dir="io", conn=("JM2", 0))),
                Subsignal("oe", PinsN("41", dir="o", conn=("JM2", 0))),
                Subsignal("read", PinsN("43", dir="o", conn=("JM2", 0))),
                Subsignal("write", PinsN("47", dir="o", conn=("JM2", 0))),
                Subsignal("siwu", PinsN("46", dir="o", conn=("JM2", 0))),
                Subsignal("rxf", PinsN("42", dir="i", conn=("JM2", 0))),
                Subsignal("txe", PinsN("49", dir="i", conn=("JM2", 0))),
                Subsignal("wakeup", PinsN("37", dir="io", conn=("JM2", 0))),
                Subsignal("clk", Pins("40", dir="i", conn=("JM2", 0)), Clock(100e63)),

                Attrs(IOSTANDARD="LVCMOS33"),
            ),
        ])
        for i, p in enumerate([59, 61, 63, 65, 85, 87, 89, 91]):
            self.add_resources([Resource("led", i + 1, Pins(str(p), dir='o', invert=True, conn=("JM2", 0)), Attrs(IOSTANDARD="LVCMOS18"))])

        # 'source [find bus/ftdi/digilent_jtag_hs3.cfg];'
        # 'source [find cpld/xilinx-xc7.cfg]; transport select jtag;'
        # 'adapter_khz 20000;'
        # 'init;'



# An experiment that glues everything together and tries to get a full sensor -> hdmi flow working on the micro
import os
from nmigen import *
from nap import *


class Top(Elaboratable):
    def __init__(self):
        self.sensor_reset_n = ControlSignal(name='sensor_reset', reset=1)

    def elaborate(self, platform):
        from nap.platform.plugins.hdmi_plugin_resource import hdmi_plugin_connect
        hdmi_plugin_connect(platform, "north")

        m = Module()

        platform.ps7.fck_domain(143e6, "axi_hp")

        # Control Pane
        i2c_pads = platform.request("i2c")
        m.submodules.i2c = BitbangI2c(i2c_pads)

        # Input Pipeline
        sensor = platform.request("sensor")
        platform.ps7.fck_domain(24e6, "sensor_clk")
        m.d.comb += sensor.clk.eq(ClockSignal("sensor_clk"))
        m.d.comb += sensor.reset.eq(~self.sensor_reset_n)
        # TODO: find more idiomatic way to do this
        os.environ["NMIGEN_add_constraints"] = \
            "set_property CLOCK_DEDICATED_ROUTE FALSE [get_nets pin_sensor_0__lvds_clk/hispi_sensor_0__lvds_clk__i]"

        ip = Pipeline(m, prefix="input")
        ip += HispiRx(sensor, hispi_domain="hispi")

        ip += BufferedAsyncStreamFIFO(ip.output, 2048, o_domain="axi_hp")
        ip += StreamResizer(ip.output, 64)
        ip += ImageStream2PacketizedStream(ip.output)
        ip += DramPacketRingbufferStreamWriter(ip.output, max_packet_size=0x800000, n_buffers=4)
        dram_writer = ip.last
        ip += DramPacketRingbufferCpuReader(dram_writer)

        # Output pipeline
        op = Pipeline(m, prefix="output", start_domain="axi_hp")
        op += DramPacketRingbufferStreamReader(dram_writer)
        op += StreamResizer(op.output, 48)
        op += StreamGearbox(op.output, target_width=12)
        op += PacketizedStream2ImageStream(op.output, 2304)
        op += StreamResizer(op.output, 8, upper_bits=True)
        op += StreamBuffer(op.output)
        op += SimpleInterpolatingDebayerer(op.output, 2304, 1296)
        op += StreamBuffer(op.output)
        op += VideoResizer(op.output, 2560, 1440)

        op += BufferedAsyncStreamFIFO(op.output, depth=32 * 1024, o_domain="pix")

        hdmi = platform.request("hdmi", "north")
        op += HdmiStreamSink(op.output, hdmi, generate_modeline(2560, 1440, 30), pix_domain="pix")  # TODO: the domain handling here stinks

        return m

    @driver_method
    def kick_sensor(self):
        from os import system
        system("cat /axiom-api/scripts/kick/value")


if __name__ == "__main__":
    cli(Top, runs_on=(MicroR2Platform,), possible_socs=(ZynqSocPlatform,))



from itertools import chain
from nmigen import *
from nap import ControlSignal, nAbsDifference
from . import ImageStream, RGB24, ImageConvoluter

__all__ = ["FocusPeeking"]


class FocusPeeking(Elaboratable):
    """Adds A focus peeking overlay to the image"""

    def __init__(self, input: ImageStream, width=3000, height=3000):
        self.input = input
        self.output = ImageStream(24)

        self.width = width
        self.height = height

        self.threshold = ControlSignal(16, reset=255)
        self.highlight_r = ControlSignal(8, reset=255)
        self.highlight_g = ControlSignal(8)
        self.highlight_b = ControlSignal(8)

    def elaborate(self, platform):
        m = Module()

        def transformer_function(x, y, image_proxy):
            self_rgb = RGB24(image_proxy[x, y])
            other_rgbs = [
                RGB24(image_proxy[x + dx, y + dy])
                for dx in range(-1, 2)
                for dy in range(-1, 2)
            ]

            deviations = [[nAbsDifference(self_rgb.r, o.r), nAbsDifference(self_rgb.g, o.g), nAbsDifference(self_rgb.b, o.b)] for o in other_rgbs]
            total_deviation = sum(chain(*deviations))

            output = RGB24()
            m.d.comb += output.eq(RGB24(image_proxy[x, y]))
            with m.If(total_deviation > self.threshold):
                m.d.comb += output.r.eq(self.highlight_r)
                m.d.comb += output.g.eq(self.highlight_g)
                m.d.comb += output.b.eq(self.highlight_b)

            return output

        video_transformer = m.submodules.video_transformer = ImageConvoluter(self.input, transformer_function,
                                                                              self.width, self.height)
        m.d.comb += self.output.connect_upstream(video_transformer.output)

        return m


if __name__ == "__main__":
    in_stream = ImageStream(len(RGB24()))
    dut = FocusPeeking(in_stream, width=512, height=512)

    from nmigen.back.cxxrtl import convert
    from pathlib import Path

    cpp_path = (Path(__file__).parent / "focus_peeking_cxxrtl_test" / "focus_peeking_test.cpp")
    cpp_path.write_text(
        convert(dut, ports=[*in_stream.signals, *dut.output.signals, dut.threshold, dut.highlight_r, dut.highlight_g, dut.highlight_b])
    )



from collections import OrderedDict
from functools import lru_cache
from os.path import join, dirname
from nmigen import *
from nmigen.build import Clock
from nap import DOWNWARDS, FatbitstreamContext, max_error_freq
from nap.cores import AxiEndpoint
from ..instance_helper import InstanceHelper
from .clocking import BufG

__all__ = ["PS7"]


class PS7(Elaboratable):
    def __init__(self, here_is_the_only_place_that_instanciates_ps7=False):
        assert here_is_the_only_place_that_instanciates_ps7
        self.m = Module()
        self.instance = InstanceHelper("+/xilinx/cells_xtra.v", "PS7")()
        self.clock_constraints = OrderedDict()

    def _axi_slave_helper(self, axi, ps7_port):
        m = self.m
        m.d.comb += [ps7_port.araddr.eq(axi.read_address.payload)]
        m.d.comb += [ps7_port.arvalid.eq(axi.read_address.valid)]
        m.d.comb += [ps7_port.arid.eq(axi.read_address.id)]
        m.d.comb += [ps7_port.arburst.eq(axi.read_address.burst_type)]
        m.d.comb += [ps7_port.arl.en.eq(axi.read_address.burst_len)]
        m.d.comb += [ps7_port.arsize.eq(axi.read_address.beat_size_bytes)]
        m.d.comb += [ps7_port.arprot.eq(axi.read_address.protection_type)]
        m.d.comb += [axi.read_address.ready.eq(ps7_port.arready)]

        m.d.comb += [axi.read_data.payload.eq(ps7_port.rdata)]
        m.d.comb += [axi.read_data.resp.eq(ps7_port.rre.sp)]
        m.d.comb += [axi.read_data.valid.eq(ps7_port.rvalid)]
        m.d.comb += [axi.read_data.id.eq(ps7_port.rid)]
        m.d.comb += [axi.read_data.last.eq(ps7_port.rlast)]
        m.d.comb += [ps7_port.rre.ady.eq(axi.read_data.ready)]

        m.d.comb += [ps7_port.awaddr.eq(axi.write_address.payload)]
        m.d.comb += [ps7_port.awvalid.eq(axi.write_address.valid)]
        m.d.comb += [ps7_port.awid.eq(axi.write_address.id)]
        m.d.comb += [ps7_port.awburst.eq(axi.write_address.burst_type)]
        m.d.comb += [ps7_port.awl.en.eq(axi.write_address.burst_len)]
        m.d.comb += [ps7_port.awsize.eq(axi.write_address.beat_size_bytes)]
        m.d.comb += [ps7_port.awprot.eq(axi.write_address.protection_type)]
        m.d.comb += [axi.write_address.ready.eq(ps7_port.awready)]

        m.d.comb += [ps7_port.wdata.eq(axi.write_data.payload)]
        m.d.comb += [ps7_port.wvalid.eq(axi.write_data.valid)]
        m.d.comb += [ps7_port.wstrb.eq(axi.write_data.byte_strobe)]
        m.d.comb += [ps7_port.wid.eq(axi.write_data.id)]
        m.d.comb += [ps7_port.wlast.eq(axi.write_data.last)]
        m.d.comb += [axi.write_data.ready.eq(ps7_port.wready)]

        m.d.comb += [axi.write_response.resp.eq(ps7_port.bre.sp)]
        m.d.comb += [axi.write_response.valid.eq(ps7_port.bvalid)]
        m.d.comb += [axi.write_response.id.eq(ps7_port.bid)]
        m.d.comb += [ps7_port.bre.ady.eq(axi.write_response.ready)]

        # we add the non standard fifo levels to the axi interface because we need
        # them to reset the reader / writer cores when they are stuck (e.g. after
        # reloading the PL
        axi.read_address_fifo_level = Signal(3) @ DOWNWARDS
        m.d.comb += axi.read_address_fifo_level.eq(ps7_port.racount)
        axi.read_data_fifo_level = Signal(8) @ DOWNWARDS
        m.d.comb += axi.read_data_fifo_level.eq(ps7_port.rcount)
        axi.write_address_fifo_level = Signal(6) @ DOWNWARDS
        m.d.comb += axi.write_address_fifo_level.eq(ps7_port.wacount)
        axi.write_data_fifo_level = Signal(8) @ DOWNWARDS
        m.d.comb += axi.write_data_fifo_level.eq(ps7_port.wcount)

    def _axi_master_helper(self, axi, ps7_port):
        m = self.m
        m.d.comb += [axi.read_address.payload.eq(ps7_port.araddr)]
        m.d.comb += [axi.read_address.valid.eq(ps7_port.arvalid)]
        m.d.comb += [axi.read_address.id.eq(ps7_port.arid)]
        m.d.comb += [axi.read_address.burst_type.eq(ps7_port.arburst)]
        m.d.comb += [axi.read_address.burst_len.eq(ps7_port.arl.en)]
        m.d.comb += [axi.read_address.beat_size_bytes.eq(ps7_port.arsize)]
        m.d.comb += [axi.read_address.protection_type.eq(ps7_port.arprot)]
        m.d.comb += [ps7_port.arready.eq(axi.read_address.ready)]

        m.d.comb += [ps7_port.rdata.eq(axi.read_data.payload)]
        m.d.comb += [ps7_port.rre.sp.eq(axi.read_data.resp)]
        m.d.comb += [ps7_port.rvalid.eq(axi.read_data.valid)]
        m.d.comb += [ps7_port.rid.eq(axi.read_data.id)]
        m.d.comb += [ps7_port.rlast.eq(axi.read_data.last)]
        m.d.comb += [axi.read_data.ready.eq(ps7_port.rre.ady)]

        m.d.comb += [axi.write_address.payload.eq(ps7_port.awaddr)]
        m.d.comb += [axi.write_address.valid.eq(ps7_port.awvalid)]
        m.d.comb += [axi.write_address.id.eq(ps7_port.awid)]
        m.d.comb += [axi.write_address.burst_type.eq(ps7_port.awburst)]
        m.d.comb += [axi.write_address.burst_len.eq(ps7_port.awl.en)]
        m.d.comb += [axi.write_address.beat_size_bytes.eq(ps7_port.awsize)]
        m.d.comb += [axi.write_address.protection_type.eq(ps7_port.awprot)]
        m.d.comb += [ps7_port.awready.eq(axi.write_address.ready)]

        m.d.comb += [axi.write_data.payload.eq(ps7_port.wdata)]
        m.d.comb += [axi.write_data.valid.eq(ps7_port.wvalid)]
        m.d.comb += [axi.write_data.byte_strobe.eq(ps7_port.wstrb)]
        m.d.comb += [axi.write_data.id.eq(ps7_port.wid)]
        m.d.comb += [axi.write_data.last.eq(ps7_port.wlast)]
        m.d.comb += [ps7_port.wready.eq(axi.write_data.ready)]

        m.d.comb += [ps7_port.bre.sp.eq(axi.write_response.resp)]
        m.d.comb += [ps7_port.bvalid.eq(axi.write_response.valid)]
        m.d.comb += [ps7_port.bid.eq(axi.write_response.id)]
        m.d.comb += [axi.write_response.ready.eq(ps7_port.bre.ady)]

    gp_master_number = 0
    def get_axi_gp_master(self, clk) -> AxiEndpoint:
        axi = AxiEndpoint(addr_bits=32, data_bits=32, lite=False, id_bits=12)

        ps7_port = self.instance.maxigp[self.gp_master_number]
        self.gp_master_number += 1
        self.m.d.comb += ps7_port.aclk.eq(clk)
        self._axi_master_helper(axi, ps7_port)

        return axi

    hp_slave_number=0
    def get_axi_hp_slave(self, clk) -> AxiEndpoint:
        axi = AxiEndpoint(addr_bits=32, data_bits=64, lite=False, id_bits=12)

        ps7_port = self.instance.saxi.hp[self.hp_slave_number]
        self.hp_slave_number += 1
        self.m.d.comb += ps7_port.aclk.eq(clk)
        self._axi_slave_helper(axi, ps7_port)

        return axi

    @staticmethod
    @lru_cache()
    def get_possible_fclk_frequencies():
        with open(join(dirname(__file__), "ps7_fclk_frequencies.txt")) as f:
            return [int(x) for x in f.readlines()]

    def fck_domain(self, requested_frequency=100e6, domain_name="sync", max_error_percent=1) -> Clock:
        """
        Creates a clockdomain driven by a PS7 fclk
        :param max_error_percent:
        :param domain_name: teh name of the domain to create
        :param requested_frequency: the requested frequency
        :return: the real freqency that will be generated by the fclk
        """
        self.m.domains += ClockDomain(domain_name)
        fclk_num = len(self.clock_constraints)

        driving_signal = Signal(attrs={"KEEP": "TRUE"}, name="{}_driving_signal".format(domain_name))
        self.m.d.comb += driving_signal.eq(self.instance.fclk.clk[fclk_num])

        bufg = self.m.submodules["fclk_bufg_{}".format(fclk_num)] = BufG(driving_signal)

        real_freq = [x for x in self.get_possible_fclk_frequencies() if x <= requested_frequency][-1]
        max_error_freq(real_freq, requested_frequency, max_error_percent)
        self.clock_constraints[fclk_num] = (driving_signal, bufg.o, real_freq, domain_name)
        return Clock(real_freq)

    def elaborate(self, platform):
        m = Module()

        m.submodules.ps7_block = self.instance
        m.submodules.connections = self.m

        for i, (clock_signal, bufg_out, frequency, domain_name) in self.clock_constraints.items():
            if hasattr(platform, "add_clock_constraint"):
                self.m.d.comb += ClockSignal(domain_name).eq(bufg_out)
                platform.add_clock_constraint(clock_signal, frequency)
            else:  # we are a sim platform
                platform.add_sim_clock(domain_name, frequency)

        fc = FatbitstreamContext.get(platform)
        fc.init_commands.insert(
            0,  # we insert this code at the beginning of the init sequence because otherwise the zynq might hang
                # (e.g. when the clock is not setup but we try to access something via axi)
            "\n".join(
                "# clockdomain '{name}':\n"
                "echo 1 > /sys/class/fclk/fclk{i}/enable\n"
                "echo {freq} > /sys/class/fclk/fclk{i}/set_rate\n"
                    .format(freq=int(freq), i=i, name=domain_name)
                for i, (clock_signal, bufg_out, freq, domain_name) in self.clock_constraints.items()
            )
        )
        fc.init_commands.append("# set the bit width of all axi hp slaves to 64 bits")
        for base in [0xF8008000, 0xF8009000, 0xF800A000, 0xF800B000]:
            fc.init_commands.append("devmem2 0x{:x} w 0".format(base))
            fc.init_commands.append("devmem2 0x{:x} w 0xF00".format(base + 0x14))
        fc.init_commands.append("# set urgent to all axi hp slaves")
        fc.init_commands.append("devmem2 0xF8000600 w 0xcc")


        fc.init_commands.append("\n")

        return m



from nmigen import *
from nmigen.lib.fifo import SyncFIFO
from nap import nAny


class Ila(Elaboratable):
    def __init__(self, trace_length=1000, fifo_underpower=0.5):
        self.probes = []
        self.triggers = []
        self.trace_length = trace_length
        self.fifo_underpower = fifo_underpower

        self.dropped = Signal(range(trace_length))

    def probe(self, signal):
        self.probes += [signal]
        return signal

    def trigger(self, signal):
        self.triggers += [signal]
        return signal

    def elaborate(self, platform):
        m = Module()

        fifo = m.submodules.fifo_type = SyncFIFO(width=sum(len(probe) for probe in self.probes), depth=int(self.trace_length * self.fifo_underpower), fwft=False)

        current_sample = Signal.like(self.dropped)
        with m.FSM():
            with m.State("IDLE"):
                with m.If(nAny(self.triggers)):
                    m.d.sync += current_sample.eq(0)
                    m.d.sync += self.dropped.eq(0)
                    m.next = "TRIGGERED"
            with m.State("TRIGGERED"):
                m.d.comb += fifo.w_data.eq(Cat(*self.probes))
                m.d.comb += fifo.w_en.eq(1)
                
                with m.If(~fifo.w_rdy):
                    m.d.sync += self.dropped.eq(self.dropped + 1)

                with m.If(current_sample < self.trace_length):
                    m.d.sync += current_sample.eq(current_sample + 1)
                with m.Else():
                    m.next = "READOUT"
            with m.State("READOUT"):
                # we stay in this state forever, until we are reset
                # readout happens via axi
                pass

        return m



from nmigen import *
from nap import ControlSignal
from nap.vendor.xilinx_s7 import BufIO, ClockDivider, IDelay, DDRDeserializer

__all__ = ["MipiClockRxPhy", "MipiLaneRxPhy"]


class MipiClockRxPhy(Elaboratable):
    """Drives the sync domain with the word clock and produces a ddr bit clock derived from the clock lane at `pin`"""
    def __init__(self, pin, ddr_domain):
        self.pin = pin
        self.ddr_domain = ddr_domain

    def elaborate(self, platform):
        m = Module()

        m.domains += ClockDomain(self.ddr_domain)

        bufio = m.submodules.bufio = BufIO(self.pin)
        m.d.comb += ClockSignal(self.ddr_domain).eq(bufio.o)

        divider = m.submodules.divider = ClockDivider(self.pin, divider=4)
        m.d.comb += ClockSignal().eq(divider.o)
        platform.add_clock_constraint(divider.o, 350e6 / 4)

        return m


class MipiLaneRxPhy(Elaboratable):
    def __init__(self, pin, ddr_domain):
        self.pin = pin
        self.ddr_domain = ddr_domain

        self.bitslip = ControlSignal()

        self.output = Signal(8)

    def elaborate(self, platform):
        m = Module()

        delay = m.submodules.delay = IDelay(self.pin)
        serdes = m.submodules.serdes = DDRDeserializer(delay.output, self.ddr_domain, bit_width=8, msb_first=False)
        m.d.comb += serdes.bitslip.eq(self.bitslip)
        m.d.comb += self.output.eq(serdes.output)

        return m



from enum import Enum
from typing import Callable

from nmigen import *

from .memorymap import MemoryMap

__all__ = ["Peripheral", "Response"]


class Response(Enum):
    OK = 0
    ERR = 1


HandleRead = Callable[[Module, Signal, Signal, Callable[[Response], None]], None]
HandleWrite = Callable[[Module, Signal, Signal, Callable[[Response], None]], None]


class Peripheral(Elaboratable):
    def __init__(
            self,
            handle_read: HandleRead,
            handle_write: HandleWrite,
            memorymap: MemoryMap
    ):
        """
        A `Peripheral` is a thing that is memorymaped in the SOC.
        It gets collected and wired up automatically to a platform dependent `Controller` implementation (e.g. an
        `AxiLiteController`) by the concrete SOCPlatform (e.g. the `ZynqSocPlatform`).

        :param handle_read: a function with the signature handle_read(m, addr, data, read_done) that is used to insert
                            logic to the read path. Read_done is a function that gets a Response as an argument
        :param handle_write: a function with the signature handle_write(m, addr, data, write_done) that is used to
                            insert logic to the write path. Write_done is a function that gets a Response as an argument
        :param memorymap: the MemoryMap of the peripheral
        """
        self.handle_read = handle_read
        self.handle_write = handle_write
        self.memorymap = memorymap

    def range(self):
        return self.memorymap.absolute_range_of_direct_children.range()

    def elaborate(self, platform):
        m = Module()
        m.peripheral = self
        return m



from nmigen import *
from nap import StatusSignal, ControlSignal
from nap.cores import StreamInfo
from . import ImageStream

__all__ = ["VideoResizer"]


class VideoResizer(Elaboratable):
    """Resize an ImageStream by cropping it / extending it with blackness to the desired resolution"""
    def __init__(self, input: ImageStream, desired_width, desired_height):
        self.input = input
        self.output = input.clone(name="resized")

        self.output_width = ControlSignal(16, reset=desired_width)
        self.output_height = ControlSignal(16, reset=desired_height)
        self.shift_x = ControlSignal(signed(16))
        self.shift_y = ControlSignal(signed(16))

        self.input_width = StatusSignal(16)
        self.input_height = StatusSignal(16)

    def elaborate(self, platform):
        m = Module()

        input_x = Signal(16)
        input_y = Signal(16)

        input_read = (self.input.ready & self.input.valid)
        with m.If(input_read):
            with m.If(~self.input.line_last):
                m.d.sync += input_x.eq(input_x + 1)
            with m.Else():
                m.d.sync += input_x.eq(0)
                m.d.sync += self.input_width.eq(input_x + 1)
                m.d.sync += input_y.eq(input_y + 1)
            with m.If(self.input.frame_last):
                m.d.sync += input_y.eq(0)
                m.d.sync += self.input_height.eq(input_y + 1)

        output_x = Signal(16)
        output_y = Signal(16)
        output_write = (self.output.ready & self.output.valid)
        with m.If(output_write):
            with m.If(output_x < self.output_width - 1):
                m.d.sync += output_x.eq(output_x + 1)
            with m.Else():
                m.d.sync += output_x.eq(0)
                m.d.comb += self.output.line_last.eq(1)
                with m.If(output_y < self.output_height - 1):
                    m.d.sync += output_y.eq(output_y + 1)
                with m.Else():
                    m.d.sync += output_y.eq(0)
                    m.d.comb += self.output.frame_last.eq(1)

        eff_x = output_x - self.shift_x
        eff_y = output_y - self.shift_y
        with m.If((eff_x == input_x) & (eff_y == input_y)):
            m.d.comb += self.output.valid.eq(self.input.valid)
            m.d.comb += self.input.ready.eq(self.output.ready)
            m.d.comb += self.output.payload.eq(self.input.payload)
        with m.Elif(((eff_x < input_x) | (eff_y < input_x)) & (eff_x >= 0) & (eff_y >= 0)):
            m.d.comb += self.output.valid.eq(0)
            m.d.comb += self.input.ready.eq(1)
        with m.Else():
            m.d.comb += self.output.valid.eq(1)
            m.d.comb += self.input.ready.eq(0)
            m.d.comb += self.output.payload.eq(0)

        m.submodules.input_stream_info = StreamInfo(self.input)
        m.submodules.output_stream_info = StreamInfo(self.output)

        return m



import unittest
from os.path import join, dirname
import imageio
from nmigen import *
from nmigen.sim import Passive
from nap import SimPlatform
from nap.stream import write_to_stream
from nap.cores.video import ImageStream, RecoloringDebayerer, SimpleInterpolatingDebayerer, write_frame_to_stream, read_frame_from_stream, to_8bit_rgb, crop


class DebayerTest(unittest.TestCase):
    def check_output_stable(self, debayerer_gen):
        platform = SimPlatform()
        m = Module()

        input = ImageStream(8)
        transformer = m.submodules.transformer = debayerer_gen(input)
        image = imageio.imread(join(dirname(__file__), "test_bayer.png"))

        def write_process():
            yield from write_frame_to_stream(input, image, pause=False)
            yield from write_frame_to_stream(input, image, pause=False)
            yield from write_frame_to_stream(input, image, pause=False)
            yield Passive()
            while True:
                yield from write_to_stream(input, line_last=0, frame_last=0, payload=0)

        def read_process():
            (yield from read_frame_from_stream(transformer.output, timeout=1000, pause=False))
            first = crop(to_8bit_rgb((yield from read_frame_from_stream(transformer.output, timeout=1000, pause=False))), 1, 1, 1, 1)
            second = crop(to_8bit_rgb((yield from read_frame_from_stream(transformer.output, timeout=1000, pause=False))), 1, 1, 1, 1)
            imageio.imsave(platform.output_filename_base + "_first.png", first)
            imageio.imsave(platform.output_filename_base + "_second.png", second)
            self.assertEqual(first, second)

        platform.add_sim_clock("sync", 100e6)
        platform.add_process(write_process, "sync")
        platform.sim(m, read_process)

    def test_output_stable_recoloring_debayerer(self):
        self.check_output_stable(RecoloringDebayerer)

    def test_output_stable_simple_interpolating_debayerer(self):
        self.check_output_stable(lambda input: SimpleInterpolatingDebayerer(input, 70, 48))